"""This module allows to implement Domain Driven Design in Python.

More specificaly, CQRS and Event sourcing are implemented through the use
of DomainObjects and DomainEvents.

"""

__version__ = "1.3.1"
